###############################################################################
# mailbot.tcl - an irc mail system
# version 1.01
# 2006, 2007
# coded by comrade (comrade2k@hotmail.com)
# IRC: #coders on EFnet
# Web: comrade64.cjb.net / comrade.ownz.com / comrade.win32asm.com
###############################################################################

###############################################################################

namespace eval mailbot {
	source "mailbot-options.tcl"

	# database schema
	# CREATE TABLE MESSAGES (MessageID INTEGER PRIMARY KEY AUTOINCREMENT, DateSent INTEGER, Sender TEXT, Recipient TEXT, Message TEXT);
	# CREATE TABLE USERS (Username TEXT UNIQUE, RemindAlert INTEGER, JoinAlert INTEGER, Timezone INTEGER, LastRead INTEGER, Alias TEXT);
	# CREATE INDEX MESSAGES_DATE_INDEX ON MESSAGES (DateSent);

	# internal variables - do not change
	variable crypto_key ""
	variable crypto_block_length 16
	variable option_encrypt_names 1
	variable option_encrypt_msgs 1
	variable option_encrypt_misc 1

	################################################################
	# initializes the script and loads the database
	#
	# arguments:
	#	db_key - the password to the database
	#
	# returns:
	#	error message on failure, empty string otherwise
	#
	################################################################
	proc initialize {db_key} {
		variable crypto_key
		variable db_path

		# try to initialize the database
		if {[catch { ::sqlite3 maildb $db_path } err]} {
			return $err
		}

		# remember the password if we succeeded
		set crypto_key $db_key

		return ""
	}

	################################################################
	# returns whether the database is initialized or not
	#
	# arguments:
	#	none
	#
	# returns:
	#	1 if database is initialized, 0 otherwise
	#
	################################################################
	proc initialized {} {
		variable crypto_key

		return [expr [string length $crypto_key] != 0]
	}

	################################################################
	#
	# channel interface to the script
	#
	################################################################
	proc pub_mail {nick uhost handle channel arg} {
		variable pub_cmdname

		if {[llength $arg] >= 1} {
			switch [string tolower [lindex $arg 0]] {
			"help" {
   				cmd_help $nick $handle [lindex $arg 1]
			}
			"add" {
				cmd_post $nick $handle $channel [remove_args $arg 1]
			}
			"post" {
				cmd_post $nick $handle "" [remove_args $arg 1]
			}
			"send" {
				cmd_post $nick $handle [lindex $arg 1] [remove_args $arg 2]
			}
			"read" {
				cmd_read $nick $handle [remove_args $arg 1]
			}
			"search" {
				cmd_search $nick $handle [remove_args $arg 1]
			}
			"timezone" {
				cmd_timezone $nick $handle [lindex $arg 1]
			}
			"alias" {
				#cmd_alias $nick $handle [lindex $arg 1]
			}
			"status" {
				cmd_status $nick $handle
			}
			"check" {
				on_join $nick $uhost $handle $channel
			}
			default {
				# unknown command
				putserv "PRIVMSG $nick :Unknown command. Use \x2$pub_cmdname help\x2 to list available commands."
			}
			}
		}

		return 0
	}

	################################################################
	#
	# private messaging interface to the script
	#
	################################################################
	proc msg_mail {nick uhost handle arg} {
		variable pub_cmdname

		if {[llength $arg] >= 1} {
			switch [string tolower [lindex $arg 0]] {
			"help" {
				cmd_help $nick $handle [lindex $arg 1]
			}
			"add" {
				cmd_post $nick $handle [lindex $arg 1] [remove_args $arg 2]
			}
			"post" {
				cmd_post $nick $handle "" [remove_args $arg [list 0]]
			}
			"send" {
				cmd_post $nick $handle [lindex $arg 1] [remove_args $arg 2]
			}
			"read" {
				cmd_read $nick $handle [remove_args $arg 1]
			}
			"search" {
				cmd_search $nick $handle [remove_args $arg 1]
			}
			"timezone" {
				cmd_timezone $nick $handle [lindex $arg 1]
			}
			"alias" {
				#cmd_alias $nick $handle [lindex $arg 1]
			}
			"init" {
				cmd_init $nick $handle [lindex $arg 1] [remove_args $arg 2]
			}
			"status" {
				cmd_status $nick $handle
			}
			default {
				# unknown command
				putserv "PRIVMSG $nick :Unknown command. Use \x2$pub_cmdname help\x2 to list available commands."
			}
			}
		}

		return 0
	}

	################################################################
	# command: help
	#	show the help menu
	#
	# arguments:
	#	nick	- nickname of the user invoking the command
	#	handle	- handle of the user invoking the command
	#	arg		- command name for extended help
	#
	# returns:
	#	void
	#
	################################################################
	proc cmd_help {nick handle arg} {
		variable pub_cmdname

		switch [string tolower $arg] {
		"add" {
			::putserv "PRIVMSG $nick :\02add\x2 - sends mail to a channel."
			::putserv "PRIVMSG $nick :If invoked from a channel, simply use \x2add my message\x2 to send mail to the channel the command is invoked from."
			::putserv "PRIVMSG $nick :If invoked from private query, use \x2add #channel my message\x2 to send mail to the specific channel."
		}
		"send" {
			::putserv "PRIVMSG $nick :\x2send\x2 - sends mail to a user."
			::putserv "PRIVMSG $nick :Use \x2send username my message\x2 to send mail to a specific username."
		}
		"post" {
			::putserv "PRIVMSG $nick :\x2post\x2 - sends mail to all users in the system."
			::putserv "PRIVMSG $nick :Use \x2post\x2 to send mail to all users in the system - any user who has a user record on the bot will be able to read the message."
		}
		"read" {
			::putserv "PRIVMSG $nick :\x2read\x2 - displays all unread messages - global, private, and all channel messages the user has access to."
			::putserv "PRIVMSG $nick :Specify \x2-all\x2 to display all messages ever sent. Beware, this command may produce a lot of traffic."
			::putserv "PRIVMSG $nick :For other filtering options, see \x2$pub_cmdname help search\x2."
		}
		"search" {
			::putserv "PRIVMSG $nick :\x2search\x2 - searches the messages for a specific keyword or filter."
			::putserv "PRIVMSG $nick :Specify \x2-before:YYYY-MM-DD\x2 to search only messages before a specific date."
			::putserv "PRIVMSG $nick :Specify \x2-after:YYYY-MM-DD\x2 to search only messages after and including a specific date."
			::putserv "PRIVMSG $nick :Specify \x2-ago:num_days\x2 to search only messages from up to \x2num_days\x2 ago."
			::putserv "PRIVMSG $nick :Specify \x2-from:user\x2 to search only messages sent by the user \x2user\x2."
			::putserv "PRIVMSG $nick :Specify \x2-to:channel\x2 to search only messages sent to the channel \02channel\x2."
			::putserv "PRIVMSG $nick :Specify \x2-any\x2 to make the filter match any one of the keywords from the list. if not specified, all keywords must be matched."
			::putserv "PRIVMSG $nick :Then, specify a list of keywords to be matched in the mail body message. Specify none to match all messages."
			::putserv "PRIVMSG $nick : "
			::putserv "PRIVMSG $nick :For example, \x2$pub_cmdname search -before:2007-11-01 -after:2007-10-01 -from:leonidas sparta\x2 will search for all mail for the month of October 2007, sent by leonidas, and containing the keyword 'sparta'."
		}
		"timezone" {
			::putserv "PRIVMSG $nick :\x2timezone\x2 - adjust the time zone for displaying dates and times."
			::putserv "PRIVMSG $nick :Specify a time zone adjustment as +minutes or -minutes to make sure the dates and times display properly. The timezone adjustment is relative to the local time of the system the bot is running on."
			::putserv "PRIVMSG $nick :To view your current time zone setting, specify no adjustment."
			::putserv "PRIVMSG $nick : "
			::putserv "PRIVMSG $nick :For example, \x2$pub_cmdname timezone +120\x2 will add 2 hours to all displayed times."
		}
		default {
			::putserv "PRIVMSG $nick :The mail bot supports the following commands, all prefixed with \x2$pub_cmdname\x2:"
			::putserv "PRIVMSG $nick :\02add\x2 - sends mail to a channel. when invoked from a channel, simply use \02add \[message]\x2 and the message will be sent to the channel the command is invoked from. if invoking from a private query, use \02add #channel \[message]\x2 to specify both the channel and message"
			::putserv "PRIVMSG $nick :\x2send \[username] \[message]\x2 - sends mail to a user"
			::putserv "PRIVMSG $nick :\x2post \[message]\x2 - sends mail to everyone in the system"
			::putserv "PRIVMSG $nick :\x2read\x2 \[-all] \[-before:YYYY-MM-DD\] \[-after:YYYY-MM-DD] \[-ago:num_days]- reads all unread mail from all channels. specify \x2-all\x2 to read all mail ever sent"
			::putserv "PRIVMSG $nick :\x2search\x2 \[-before:YYYY-MM-DD\] \[-after:YYYY-MM-DD] \[-ago:num_days] \[-from:user] \[-to:channel] \[-any] \[keywords] - searches mail"
			::putserv "PRIVMSG $nick :\x2timezone\x2 \[new timezone] - sets the local time zone. displays the current timezone if no new timezone is specified"

			if {[superuser $handle]} {
				::putserv "PRIVMSG $nick : "
				::putserv "PRIVMSG $nick :\x2init\x2 \[db password] - initializes the database with the specified password"
				::putserv "PRIVMSG $nick :\x2status\x2 - prints the status of the script"
			}

			::putserv "PRIVMSG $nick :\x2help \[command]\x2 - shows this message, or detailed help about a command"
			::putserv "PRIVMSG $nick : "
			::putserv "PRIVMSG $nick :This is version 1.0, coded by comrade in 2007. Contact: #coders @ EFnet, e-mail comrade2k@hotmail.com, visit comrade64.cjb.net / comrade.ownz.com / comrade.win32asm.com"
		}
		}
	}

	################################################################
	#
	# checks if there is any unread mail for the user, and if so,
	# notifies the user
	#
	################################################################
	proc on_join {nick uhost handle channel} {
		putlog "$nick ($uhost) who is really $handle joined $channel"

		# are we even initialized?
		if {![initialized]} {
			return
		}

		array set useropts [user_getoptions $handle]
		set got_mail 0

		# check for global or private messages
		set erecp [encrypt_name $handle]
		if {[maildb exists {SELECT * FROM MESSAGES WHERE (Recipient=$erecp OR Recipient="") AND DateSent >= $useropts(LastRead) ORDER BY DateSent}]} {
			set got_mail 1
		}

		# if there are no unread global or private messages, check channel messages...
		if {$got_mail == 0} {
			# check the mail for channels the user is on
			foreach channel [::channels] {
				# check if the user has permissions to read mail on it
				if {[::matchattr $handle "|v" $channel]} {
					set echan [encrypt_name $channel]

					if {[maildb exists {SELECT * FROM MESSAGES WHERE Recipient=$echan AND DateSent >= $useropts(LastRead) ORDER BY DateSent}]} {
						set got_mail 1
					}
				}
			}
		}

		if {$got_mail > 0} {
			::putserv "PRIVMSG $nick :You have unread mail."
			return
		}

	}

	################################################################
	# command: post mail
	#	sends mail to a user, a channel, or everyone
	#
	# arguments:
	#	nick	- nickname of the user invoking the command
	#	handle	- handle of the user invoking the command
	#	recipient - handle of the user the mail is being sent to,
	#				or, the channel the mail is being posted to.
	#				if blank, message is posted to everyone
	#  	msg		- the message being posted
	#
	# returns:
	#	void
	#
	################################################################
	proc cmd_post {nick handle recipient msg} {
		# are we even initialized?
		if {![initialized]} {
			::putserv "PRIVMSG $nick :The database is not initialized. A bot administrator must initialize the database."
			return
		}

		# validate arguments
		if {![string length $msg]} {
			::putserv "PRIVMSG $nick :Message not sent: please specify a non-empty message."
			return
		}

		# encrypt the variables
		set esender [encrypt_name $handle]
		set erecipient [encrypt_name $recipient]
		set emsg [encrypt_msg $msg]
		set ets [encrypt_time [clock seconds]]

		# is the recipient global?
		if {[string length $recipient] == 0} {
			# it is a global message, so make sure we set the Recipient to NULL, regardless of encryption
			maildb eval {INSERT INTO MESSAGES (Sender, Recipient, Message, DateSent) VALUES($esender, "", $emsg, $ets)}
			::putserv "PRIVMSG $nick :Global message number \x2[maildb last_insert_rowid]\x2 posted."

			return 1
		}

		# is the recipient a channel?
		if {[string  range $recipient 0 0] == "#"} {
			# if it is channel, does the sender have permissions to post to it?
			if {[::matchattr $handle "|v" $recipient]} {
				# yes, post the message
				maildb eval {INSERT INTO MESSAGES (Sender, Recipient, Message, DateSent) VALUES($esender, $erecipient, $emsg, $ets)}
				::putserv "PRIVMSG $nick :Message number \x2[maildb last_insert_rowid]\x2 posted to channel \x2$recipient\x2."

				return 1
			} else {
				# no, give the user an error
				::putserv "PRIVMSG $nick :No access to \x2$recipient\x2."

				return 0
			}
		}

		# is the recipient a user?
		if {[::validuser $recipient]} {
			# make sure the case is correct
			set recipient [getuser $recipient HANDLE]
			set erecipient [encrypt_name $recipient]

			maildb eval {INSERT INTO MESSAGES (Sender, Recipient, Message, DateSent) VALUES($esender, $erecipient, $emsg, $ets)}
			::putserv "PRIVMSG $nick :Private message number \x2[maildb last_insert_rowid]\x2 sent to user \x2$recipient\x2."

			return 1
		}

		::putserv "PRIVMSG $nick :Invalid recipient \x2$recipient\x2."
		return 0
	}

	################################################################
	# command: read mail
	#	prints the user all his unread mail.
	#	unread mail consists of private messages, as well as all
	#	channel mail on each channel the user is added on
	#
	# arguments:
	#	nick	- nickname of the user invoking the command
	#	handle	- handle of the user invoking the command
	#	arg		- optional arguments
	#
	# returns:
	#	void
	#
	################################################################
	proc cmd_read {nick handle arg} {
		# are we even initialized?
		if {![initialized]} {
			::putserv "PRIVMSG $nick :The database is not initialized. A bot administrator must initialize the database."
			return
		}

		# get the user's options and remember current time
		array set useropts [user_getoptions $handle]
		set now [clock seconds]

		# parse arguments
		set minDate $useropts(LastRead)
		set maxDate 0

		foreach {i} $arg {
			if {[string equal -nocase "all" $i] || [string equal -nocase "-all" $i]} {

				::putserv "PRIVMSG $nick :Showing ALL mail ever posted."
				set minDate 0

			} elseif {[string match -nocase "-before:*" $i]} {

				set date [clock scan [string range $i 8 [string length $i]]]
				set maxDate [expr $date - $useropts(Timezone)]

			} elseif {[string match -nocase "-after:*" $i]} {

				set date [clock scan [string range $i 7 [string length $i]]]
				set minDate [expr $date - $useropts(Timezone)]

			} elseif {[string match -nocase "-ago:*" $i]} {

				set days [string range $i 5 [string length $i]]
				set minDate [expr $now - ($days * 60 * 60 * 24)]

			}
		}

		# sanity check that minDate <= maxDate
		if {$minDate >= $maxDate && $maxDate > 0} {
			::putserv "PRIVMSG $nick :Search aborted: invalid date range specified."
		}

		set printed_mail 0

		# spew out mail sent directly to the user (private messages)
		set erecp [encrypt_name $handle]
		set first 1
		maildb eval {SELECT * FROM MESSAGES WHERE Recipient = $erecp AND DateSent >= $minDate AND (DateSent < $maxDate OR $maxDate = 0) ORDER BY DateSent} mailitem {
			if {$first} {
				print_header $nick "PRIVATE MESSAGES"
				set first 0
			}
			print_mail $nick [array get mailitem] [array get useropts]
			incr printed_mail
		}

		# spew out global mail
		set first 1
		maildb eval {SELECT * FROM MESSAGES WHERE Recipient = "" AND DateSent >= $minDate AND (DateSent < $maxDate OR $maxDate = 0) ORDER BY DateSent} mailitem {
			if {$first} {
				print_header $nick "GLOBAL MESSAGES"
				set first 0
			}
			print_mail $nick [array get mailitem] [array get useropts]
			incr printed_mail
		}

		# spew out mail sent to the channels the user is on
		foreach channel [::channels] {
			# check if the user has permissions to read mail on it
			if {[::matchattr $handle "|v" $channel]} {
				set echan [encrypt_name $channel]
				set first 1

				maildb eval {SELECT * FROM MESSAGES WHERE Recipient = $echan AND DateSent >= $minDate AND (DateSent < $maxDate OR $maxDate = 0) ORDER BY DateSent} mailitem {
					if {$first} {
						print_header $nick "MESSAGES TO $channel"
						set first 0
					}
					print_mail $nick [array get mailitem] [array get useropts]
					incr printed_mail
				}
			}
		}

		if {$printed_mail == 0} {
			::putserv "PRIVMSG $nick :No unread mail."
		}

		# update the user's last read time stamp, so he doesn't read the same news again next time
		set useropts(LastRead) [encrypt_time $now]
		user_setoptions $handle [array get useropts]
	}


	################################################################
	# command: search mail
	#	searches all mail.
	#
	# arguments:
	#	nick	- nickname of the user invoking the command
	#	handle	- handle of the user invoking the command
	#	arg		- optional arguments
	#
	# returns:
	#	void
	#
	################################################################
	proc cmd_search {nick handle arg} {
		# are we even initialized?
		if {![initialized]} {
			::putserv "PRIVMSG $nick :The database is not initialized. A bot administrator must initialize the database."
			return
		}

		# get the user's options and remember current time
		array set useropts [user_getoptions $handle]
		set now [clock seconds]

		# parse arguments
		set minDate 0
		set maxDate 0
		set matchSender ""
		set matchRecipient ""
		set matchRecpIsChan 0
		set matchRecpIsUser 0
		set matchKeywords [list]
		set matchAny 0

		foreach {i} $arg {
			if {[string match -nocase "-before:*" $i]} {

				set date [clock scan [string range $i 8 [string length $i]]]
				set maxDate [expr $date - $useropts(Timezone)]

			} elseif {[string match -nocase "-after:*" $i]} {

				set date [clock scan [string range $i 7 [string length $i]]]
				set minDate [expr $date - $useropts(Timezone)]

			} elseif {[string match -nocase "-ago:*" $i]} {

				set days [string range $i 5 [string length $i]]
				set minDate [expr $now - ($days * 60 * 60 * 24)]

			} elseif {[string match -nocase "-from:*" $i]} {

				set matchSender [string range $i 6 [string length $i]]
				if {![::validuser $matchSender]} {
					::putserv "PRIVMSG $nick :Search aborted: user \x2$matchSender\x2 unknown."
					return
				}

				set matchSender [::getuser $matchSender handle]

			} elseif {[string match -nocase "-to:*" $i]} {

				#
				# for the -to filter, we can either filter by destination channel (any sender)
				# or we can filter private messages by destination recipient (sender = searcher)
				#
				set matchRecipient [string range $i 4 [string length $i]]

				set matchRecpIsChan [::validchan $matchRecipient]
				set matchRecpIsUser [::validuser $matchRecipient]

				if {!$matchRecpIsChan && !$matchRecpIsUser} {
					::putserv "PRIVMSG $nick :Search aborted: recipient \x2$matchRecipient\x2 unknown."
					return
				}

				if {$matchRecpIsChan && ![::matchattr $handle "|v" $matchRecipient]} {
					::putserv "PRIVMSG $nick :Search aborted: access to channel \x2$matchRecipient\x2 denied."
					return
				}

				#
				# make sure the recipient has proper case
				#
				if {$matchRecpIsUser} {
					set matchRecipient [getuser $matchRecipient HANDLE]
				}

			} elseif {[string equal -nocase "-any" $i]} {

				set matchAny 1

			} else {

				lappend matchKeywords $i
			}
		}

		# sanity check that minDate <= maxDate
		if {$minDate >= $maxDate && $maxDate > 0} {
			::putserv "PRIVMSG $nick :Search aborted: invalid date range specified."
		}

		::putserv "PRIVMSG $nick :Search: matching [expr $matchAny ? \"any of\" : \"all\"] the keywords in: \x2$matchKeywords\x2"

		set printed_mail 0

		if {$matchRecpIsChan} {
			# search only mail sent to a specific channel (access previously verified)

			set echan [encrypt_name $matchRecipient]
			set esender [encrypt_name $matchSender]
			set first 1

			maildb eval {SELECT * FROM MESSAGES WHERE Recipient = $echan AND (Sender = $esender OR $matchSender = '') AND DateSent >= $minDate AND (DateSent < $maxDate OR $maxDate = 0) ORDER BY DateSent} mailitem {
				if {[match_mail [array get mailitem] $matchAny $matchKeywords]} {
					if {$first} {
						print_header $nick "MESSAGES TO $matchRecipient"
						set first 0
					}
					print_mail $nick [array get mailitem] [array get useropts]
					incr printed_mail
				}
			}

		} elseif {$matchRecpIsUser} {
			# search only mail sent to a specific user (sent by the searcher)

			set erecp [encrypt_name $matchRecipient]
			set esender [encrypt_name $handle]
			set first 1

			maildb eval {SELECT * FROM MESSAGES WHERE Recipient = $erecp AND Sender = $esender AND DateSent >= $minDate AND (DateSent < $maxDate OR $maxDate = 0) ORDER BY DateSent} mailitem {
				if {[match_mail [array get mailitem] $matchAny $matchKeywords]} {
					if {$first} {
						print_header $nick "PRIVATE MESSAGES FROM $handle TO $matchRecipient"
						set first 0
					}
					print_mail $nick [array get mailitem] [array get useropts]
					incr printed_mail
				}
			}

		} else {
			# search all mail

			set esender [encrypt_name $matchSender]

			# spew out mail sent directly to the user (private messages)
			set erecp [encrypt_name $handle]
			set first 1
			maildb eval {SELECT * FROM MESSAGES WHERE Recipient = $erecp AND (Sender = $esender OR $matchSender = '') AND DateSent >= $minDate AND (DateSent < $maxDate OR $maxDate = 0) ORDER BY DateSent} mailitem {
				if {[match_mail [array get mailitem] $matchAny $matchKeywords]} {
					if {$first} {
						print_header $nick "PRIVATE MESSAGES"
						set first 0
					}
					print_mail $nick [array get mailitem] [array get useropts]
					incr printed_mail
				}
			}

			# spew out global mail
			set first 1
			maildb eval {SELECT * FROM MESSAGES WHERE Recipient = "" AND (Sender = $esender OR $matchSender = '') AND DateSent >= $minDate AND (DateSent < $maxDate OR $maxDate = 0) ORDER BY DateSent} mailitem {
				if {[match_mail [array get mailitem] $matchAny $matchKeywords]} {
					if {$first} {
						print_header $nick "GLOBAL MESSAGES"
						set first 0
					}
					print_mail $nick [array get mailitem] [array get useropts]
					incr printed_mail
				}
			}

			# spew out mail sent to the channels the user is on
			foreach channel [::channels] {
				# check if the user has permissions to read mail on it
				if {[::matchattr $handle "|v" $channel]} {
					set echan [encrypt_name $channel]
					set first 1

					maildb eval {SELECT * FROM MESSAGES WHERE Recipient = $echan AND (Sender = $esender OR $matchSender = '') AND DateSent >= $minDate AND (DateSent < $maxDate OR $maxDate = 0) ORDER BY DateSent} mailitem {
						if {[match_mail [array get mailitem] $matchAny $matchKeywords]} {
							if {$first} {
								print_header $nick "MESSAGES TO $channel"
								set first 0
							}
							print_mail $nick [array get mailitem] [array get useropts]
							incr printed_mail
						}
					}
				}
			}
		}

		if {$printed_mail == 0} {
			::putserv "PRIVMSG $nick :No results found."
		}

		# update the user's last read time stamp, so he doesn't read the same news again next time
		set useropts(LastRead) [encrypt_time $now]
		user_setoptions $handle [array get useropts]
	}

	################################################################
	# command: view/change timezone
	#	show's the user current timezone, or changes it
	#
	# arguments:
	#	nick	- nickname of the user invoking the command
	#	handle	- handle of the user invoking the command
	#	newTimezone - optionally, the new timezone to set
	#
	# returns:
	#	void
	#
	################################################################
	proc cmd_timezone {nick handle newTimezone} {
		# are we even initialized?
		if {![initialized]} {
			::putserv "PRIVMSG $nick :The database is not initialized. A bot administrator must initialize the database."
			return
		}

		array set useropts [user_getoptions $handle]

		::putserv "PRIVMSG $nick :Current timezone: \x2[withsign $useropts(Timezone)] minutes\x2"

		if {[string length $newTimezone] > 0} {
			set useropts(Timezone) $newTimezone
			user_setoptions $handle [array get useropts]
			::putserv "PRIVMSG $nick :New timezone: \x2$useropts(Timezone)\x2 minutes"
		}

		::putserv "PRIVMSG $nick :Your local time should be: \x2[clock format [expr [clock seconds] + (60 * $useropts(Timezone))] -format %H:%M:%S]\x2"
	}

	################################################################
	# command: set alias
	#	shows or changes the user's alias in the mail system
	#
	# arguments:
	#	nick	- nickname of the user invoking the command
	#	handle	- handle of the user invoking the command
	#	newAlias - new alias for the user, optionally
	#
	# returns:
	#	void
	#
	################################################################
	proc cmd_alias {nick handle newAlias} {
		# are we even initialized?
		if {![initialized]} {
			::putserv "PRIVMSG $nick :The database is not initialized. A bot administrator must initialize the database."
			return
		}

		array set useropts [user_getoptions $handle]

		if {[string length $useropts(Alias)]} {
			::putserv "PRIVMSG $nick :Current alias: \x2$useropts(Alias)\x2"
		} else {
			::putserv "PRIVMSG $nick :No alias currently set. Using the default handle: \x2$handle\x2"
		}

		if {[string length $newAlias]} {
			if {[string equal -nocase $newAlias "-"]} {
				set oldAlias $useropts(Alias)
				set useropts(Alias) ""
				user_setoptions $handle [array get useropts]
				::putserv "PRIVMSG $nick :Alias \x2$oldAlias\x2 removed. From now on, using the default handle: \x2$handle\x2"
			} else {
				set useropts(Alias) $newAlias
				user_setoptions $handle [array get useropts]
				::putserv "PRIVMSG $nick :New alias: \x2$useropts(Alias)\x2"
			}
		}
	}

	################################################################
	# command: initialize the script and load the database
	#	loads the database and initializes the rest of the script
	#
	# arguments:
	#	nick	- nickname of the user invoking the command
	#	handle	- handle of the user invoking the command
	#	db_key_checksum - the checksum of the key
	#	db_key	- the database password
	#
	# returns:
	#	void
	#
	################################################################
	proc cmd_init {nick handle db_key_checksum db_key} {
		# validate access
		if {![superuser $handle]} {
			::putserv "PRIVMSG $nick :Access denied."
			return
		}

		# validate arguments
		if {![string length $db_key] || ![string length $db_key_checksum]} {
			::putserv "PRIVMSG $nick :Database not initialized: please specify a non-empty password and non-empty checksum."
			return
		}

		# validate checksum
		set correct_checksum [password_checksum $db_key]
		if {$db_key_checksum != $correct_checksum} {
			::putserv "PRIVMSG $nick :Database not initialized: checksums do not match. The correct checksum for \x2$db_key\x2 is \x2$correct_checksum\x2."
			::putserv "PRIVMSG $nick :\x2NOTE\x2: Initializing the database with the incorrect password can inflict \x1Firreparable damages\x1F. Please double-check the password or the checksum for typos."
			return
		}

		# try to initialize
		set err [initialize $db_key]
		if {![string length $err]} {
			::putserv "PRIVMSG $nick :Database initialized."
		} else {
			::putserv "PRIVMSG $nick :Error initializing database: $err."
		}
	}

	################################################################
	# command: status
	#	prints the status of the script
	#
	# arguments:
	#	nick	- nickname of the user invoking the command
	#	handle	- handle of the user invoking the command
	#	db_key	- the database password
	#
	# returns:
	#	void
	#
	################################################################
	proc cmd_status {nick handle} {
		# validate access
		if {![superuser $handle]} {
			::putserv "PRIVMSG $nick :Access denied."
		}

		::putserv "PRIVMSG $nick :Database initialized: \x2[initialized]\x2"
	}

	################################################################
	# returns whether the user has access to superuser commands of
	# the script
	#
	# arguments:
	#	handle	- handle of the user
	#
	# returns:
	#	1 if user is a superuser, 0 otherwise
	#
	################################################################
	proc superuser {handle} {
		return [expr [::matchattr $handle "m"] || [::matchattr $handle "n"]]
	}

	################################################################
	# prints a padded header
	#
	# arguments:
	#	nick	- the recipient
	#	header	- what to print
	#
	# returns:
	#	void
	#
	################################################################
	proc print_header {nick header} {
		set barlength 40
		set leftpad 4
		set rightpad [expr $barlength - $leftpad - 1 - [string length $header] - 1]
		set padchar " "
		::putserv "PRIVMSG $nick :\x16[string repeat $padchar $leftpad]\x2$header\x2[string repeat $padchar $rightpad]\x16"
	}

	################################################################
	# prints a mail message
	#
	# arguments:
	#	nick	- the recipient
	#	mailitem - the message
	#	_useropts - can't remember... wrote the code a long time ago
	#
	# returns:
	#	void
	#
	################################################################
	proc print_mail {nick mailitem _useropts} {
		array set mail $mailitem
		array set useropts $_useropts

		set s "\x2$mail(MessageID)\x2. "
		append s [clock format [expr [decrypt_time $mail(DateSent)] + (60 * $useropts(Timezone))] -format %d.%m.%Y\x20%H:%M:%S]
		append s " \x2[decrypt_name $mail(Sender)]\x2: "
		append s [decrypt_msg $mail(Message)]

		::putserv "PRIVMSG $nick :$s"
	}

	################################################################
	# matches the mail by given keywords
	#
	# arguments:
	#	mailitem	- the mail item in question
	#	matchAny	- true to match ANY one of the keywords,
	#				  false to match ALL of the keywords
	#	matchKeywords - the keywords to match. an empty list matches
	#					all the mail items
	#
	# returns:
	#	true if the mail item matches the keywords, false otherwise
	#
	################################################################
	proc match_mail {mailitem matchAny matchKeywords} {
		array set mail $mailitem

		# no keywords match everything
		if {[llength $matchKeywords] == 0} {
			return 1
		}

		# default match value depends on whether we are matching ANY
		# or ALL keywords
		set match [expr $matchAny ^ 1]
		set msg [decrypt_msg $mail(Message)]

		foreach {keyword} $matchKeywords {
			set m [expr [string first [string tolower $keyword] [string tolower $msg]] >= 0]
			if {$matchAny && $m} {
				set match 1
				break
			} elseif {!$matchAny && !$m} {
				set match 0
				break
			}
		}

		return $match
	}

	################################################################
	# returns a list of user's options
	#
	# arguments:
	#	username - the username of the user in the database
	#
	# returns:
	#	list of user's options
	#
	################################################################
	proc user_getoptions {username} {
		set euser [encrypt_name $username]

		# default options
		array set useropts {
			LastRead	0
			Timezone	0
			JoinAlert	0
			RemindAlert	0
			Alias		""
		}

		# read options from db, if they exist
		maildb eval {SELECT * FROM USERS WHERE Username=$euser} useropts {
		}
		unset useropts(*)
		if {[info exists useropts(Username)]} {
			unset useropts(Username)
		}

		if {[info exists useropts(Alias)]} {
			if {[string length $useropts(Alias)]} {
				set useropts(Alias) [decrypt_misc $useropts(Alias)]
			}
		}

		return [array get useropts]
	}


	################################################################
	# update's the user's options in the database.
	# if no record exists, one is created.
	#
	# arguments:
	#	username - the username of the user in the database
	#	useropts - the list of user's options
	#
	# returns:
	#	void
	#
	################################################################
	proc user_setoptions {username useropts} {
		set euser [encrypt_name $username]

		# save options to db
		set names ""
		set values ""
		foreach {optName optValue} $useropts {
			# add commas if needed
			if {[string length $names] > 0} {
				append names ", "
			}
			if {[string length $values] > 0} {
				append values ", "
			}

			# append name and value to the sql query string
			append names $optName

			# check if value is a number or a string
			if {[string is integer -strict $optValue]} {
				append values $optValue
			} else {
				append values "\"[encrypt_misc $optValue]\""
			}
		}
		set sql "INSERT OR REPLACE INTO USERS (Username, $names) VALUES(\"$euser\", $values)"
		maildb eval $sql
	}


	################################################################
	#
	# encryption layer
	#
	################################################################
	proc encrypt_name {name} {
		variable option_encrypt_names

		if {$option_encrypt_names == 1} {
			return [encrypt $name]
		} else {
			return $name
		}
	}

	proc decrypt_name {name} {
		variable option_encrypt_names

		if {$option_encrypt_names == 1} {
			return [decrypt $name]
		} else {
			return $name
		}
	}

	proc encrypt_msg {msg} {
		variable option_encrypt_msgs

		if {$option_encrypt_msgs == 1} {
			return [encrypt $msg]
		} else {
			return $msg
		}
	}

	proc decrypt_msg {msg} {
		variable option_encrypt_msgs

		if {$option_encrypt_msgs == 1} {
			return [decrypt $msg]
		} else {
			return $msg
		}
	}

	proc encrypt_time {timestamp} {
		# not supported yet
		variable option_encrypt_misc

		if {$option_encrypt_misc == 1} {
			return $timestamp
		} else {
			return $timestamp
		}
	}

	proc decrypt_time {timestamp} {
		# not supported yet
		variable option_encrypt_misc

		if {$option_encrypt_misc == 1} {
			return $timestamp
		} else {
			return $timestamp
		}
	}

	proc encrypt_misc {value} {
		variable option_encrypt_misc

		if {$option_encrypt_misc == 1} {
			return [encrypt $value]
		} else {
			return $value
		}
	}

	proc decrypt_misc {value} {
		variable option_encrypt_misc

		if {$option_encrypt_misc == 1} {
			return [decrypt $value]
		} else {
			return $value
		}
	}

	################################################################
	#
	# raw crypto routines
	#
	################################################################
	proc encrypt {input} {
		variable crypto_block_length
		
		# add the length of the string to the string itself
		set prepared_input [binary format s1a* [string length $input] $input]

		# pad up the string to the block length
		set remainder [expr [string length $prepared_input] % $crypto_block_length]
		if {$remainder > 0} {
			set prepared_input [binary format a*a* $prepared_input [string repeat "\x0" [expr $crypto_block_length - $remainder]]]
		}

		# encrypt and return
		set nilblock [string repeat \0 16]
		set key [::aes::Init cbc [prepare_key] $nilblock]
		set ciphertext [::aes::Encrypt $key $prepared_input]
		::aes::Final $key

		return [::base64::encode $ciphertext]
	}

	proc decrypt {input} {
		# decrypt
		set raw_decrypted [::aes::aes -mode cbc -dir decrypt -key [prepare_key] [::base64::decode $input]]

		# get the actual length of the text out
		binary scan $raw_decrypted s1a* actual_length decrypted

		# trim and return
		return [string range $decrypted 0 [expr $actual_length - 1]]
	}

	proc prepare_key {} {
		variable crypto_key
		variable crypto_block_length

		# spread and repeat the cryptokey if it's too small, truncate otherwise
		set proper_cryptokey [string repeat $crypto_key [expr ($crypto_block_length + [string length $crypto_key] - 1) / [string length $crypto_key]]]
		set proper_cryptokey [string range $proper_cryptokey 0 [expr $crypto_block_length - 1]]
		return [binary format a* $proper_cryptokey]
	}

	################################################################
	#
	# utility functions
	#
	################################################################

	################################################################
	# returns the sign of the value
	#
	# arguments:
	#	value	- the value
	#
	# returns:
	#	+ if the value is non-negative, - otherwise
	#
	################################################################
	proc signof {value} {
		if {$value < 0} {
			return "-"
		} else {
			return "+"
		}
	}

	################################################################
	# returns the value with the sign prefixed
	#
	# arguments:
	#	value	- the value
	#
	# returns:
	#	"+value" if the value is non-negative, "-value" otherwise
	#
	################################################################
	proc withsign {value} {
		return [join [list [signof $value] $value] ""]
	}

	################################################################
	# returns an 8-bit checksum of a password
	# the actual checksum is CRC32, with the 4 octets xorred
	# together to produce an 8-bit output
	#
	# arguments:
	#	password	- password string to be checksumed
	#
	# returns:
	#	the checksum
	#
	################################################################
	proc password_checksum {password} {
		set crc [::crc::crc32 $password]

		return [expr (($crc >> 24) & 0xFF) ^ (($crc >> 16) & 0xFF) ^ (($crc >>	8) & 0xFF) ^ (($crc >>	0) & 0xFF)]
	}
	
	################################################################
	# removes arguments from the front of the string, and returns
	# the new string
	#
	# arguments:
	#	s		- the string to have its arguments removed
	#	n  		- how many arguments to remove from the front
	#
	# returns:
	#	the string with the given arguments removed
	#
	################################################################
	proc remove_args {s n} {
		set i 0

		while {$i >= 0 && $n > 0} {
			set j [string first " " $s $i]
			if {$j < 0} {
				return ""
			}

			set i [expr $j + 1]
			set n [expr $n - 1]
		}

		return [string trimleft [string range $s $i [string length $s]]]
	}

	################################################################
	# registers bindings for eggdrop (msg, pub, join)
	################################################################
	proc initialize_eggdrop_bindings {} {
		variable pub_cmdname

		::bind pub - $pub_cmdname mailbot::pub_mail
		::bind msg - $pub_cmdname mailbot::msg_mail
		::bind join - * mailbot::on_join
	}

}
###############################################################################

# register eggdrop bindings
mailbot::initialize_eggdrop_bindings

putlog "mailbot v1.01 loaded"

###############################################################################
